#include <stack>
#include <map>
#include <algorithm>
#include <iostream>
#include "menumiso.h"
#include "mpredexp.h"

typedef pair<int,int> mp;
bool mp_comparator(const mp& l, const mp& r){
  return l.first<r.first;
}

EnumParameter::EnumParameter(){
  maxniso=10;
  numonly=false;
  debug=false;
  immreturn=false;
  minjuncfrac=0.1;
  minadjjuncfrac=2;
  minintronretentioncvgfrac=0.8;
  minintronretentionlen=10;


  RDSUPPORT_ALPHA=0.1;
  MAX_SUPP_CONSTRAINT=10000;

  assemble_by_sample=false;
  current_sample=-1;

  JS_MIN_DISTRIB_DIST=2.0;

  JS_MIN_EX_DIST=2.0;
  JS_MIN_CO_DIST=1.0;

  JS_DISTANT_EXON_ONLY=true;

  JS_MIN_SEG_LEN_TO_CONSIDER=20;

  
  DO_CORRELATION=true;
}

/*
For a given segment c, select its possible children
*/
void selectchildren(MInst& inst, int c,
  mi& rcount,vi& children,EnumParameter& para){
  float maxj=0.0;//maximum junction count
  float maxnonjj=0.0; //the maximum junction count for non-adjacent junctions
  int nsegs=rcount.size();
  int ncandidate=0;
  for(int i=c+1;i<nsegs;i++){
    if(rcount[c][i]>maxj) 
      maxj=rcount[c][i];
    if(i>c+1 && rcount[c][i]>maxnonjj)
      maxnonjj=rcount[c][i];
    if(rcount[c][i]>0)ncandidate++;
  }
  vector<mp>sch;
  if(para.debug)cout<<"Children of "<<c<<":   ";
  bool jumpretention=false;
  for(int i=c+1;i<nsegs;i++){
   if(rcount[c][i]>0 ){
     if(para.debug)cout<<i<<":"<<rcount[c][i]<<" ";
     if(i==c+1){ 
       //possible intron retention
       float mfrac=para.minjuncfrac*para.minadjjuncfrac;
       if(mfrac>1)mfrac=1;
       double cvgc=getVectorSum<double>(inst.segcvg[c]);
       double cvgn=getVectorSum<double>(inst.segcvg[i]);
       int ccount=rcount[c][i];
       //to handle very short introns (usually done by splicing errors), check the junction reads of the next node
       int cnextcount=0;
       for(int j=c+3;j<nsegs;j++){
         if(rcount[i][j]>cnextcount)cnextcount=rcount[i][j];
       }
       bool isaccepted=false;
       if(ncandidate==1) isaccepted=true;
       //cout<<c<<"->"<<i<<", c="<<ccount<<", cn="<<cnextcount<<", maxj="<<maxj<<",frac="<<mfrac<<endl;
       if(ccount>=maxj*mfrac ){
         if(inst.getSegLen(i)< para.minintronretentionlen){ 
           isaccepted=true;
           if(cnextcount>maxnonjj*mfrac){
           }
         }else{ 
           if(cvgn>=cvgc*para.minintronretentioncvgfrac ){
             isaccepted=true;
           }
         }
       }
       if(isaccepted){
          sch.push_back(mp(rcount[c][i],i));
       }
       else
          jumpretention=true;
     }else{ 
       //junctions
       float tmaxj=maxj;
       if(jumpretention) tmaxj=maxnonjj;
       if(rcount[c][i]>=tmaxj*para.minjuncfrac){ 
          sch.push_back(mp(rcount[c][i],i));
       }else{
          if(rcount[c][i]>=maxnonjj*para.minjuncfrac){
            sch.push_back(mp(rcount[c][i],i));
          }
       }
     }//end i==c+1
    }//end if 
  }//end for
  if(para.debug)cout<<endl;
  sort(sch.begin(),sch.end(),mp_comparator);
  //set up children, from largest to smallest
  for(int i=sch.size()-1;i>=0;i--) children.push_back(sch[i].second);
}


/*
Enumeration algorithm. The parameters are specified in para.
If para.immreturn is set true, no results will be given.
*/
long goenum_numonly(MInst& inst, mi & rcount, 
  int start, //start node
  mi& res,  //results stored, if numonly=false
  EnumParameter & para
  ) 
{
  
  int nsegs=inst.getNSegs();
  stack<int> cstack;
  cstack.push(start);
  
  //stores the mark of the number of childrens for current node
  vector<long> isvisited(nsegs,-1);
  //stores the current path
  vi path;
  //stores all the children of each node
  mi acceptedchildren(nsegs, vi(1,-1));
  //stores the children of the current node
  vi children;

  while(cstack.size()>0){
    int c=cstack.top();
    if(para.debug) cout<<"Node "<<c<<endl;
    children.clear();
    // set children
    if(acceptedchildren[c].size()>0 && acceptedchildren[c][0]==-1){
      selectchildren(inst,c,rcount,children,para);
      acceptedchildren[c]=children;
    }else{
      children=acceptedchildren[c];
    }

    //second visit
    if(path.size()>0 && c==path.back()){ 
      cstack.pop();
      path.pop_back();
      if(children.size()==0){
          if(isvisited[c]!=1)cerr<<"Error: not 1.\n";
      }
      else{
        //sum up all childrens
        isvisited[c]=0;
        for(int i=0;i<children.size();i++){
          int nextid=children[i];
          if(isvisited[nextid]!=-1)isvisited[c]+=isvisited[nextid];
          else cerr<<"Error: not visited .\n";
        }
      }
      if(para.immreturn && isvisited[c]>para.maxniso){
        return isvisited[c];
      }
      if(para.debug) cout<<"--Pop "<<c<<"Num:"<<isvisited[c]<<endl;
      continue;
    }else{
      //first visit
      if(para.numonly && isvisited[c]>0){
        cstack.pop();
        continue;
      }else{
        path.push_back(c);
      }
    }
    //check the children
    if(children.size()==0){
      isvisited[c]=1;
      if(!para.numonly)
        res.push_back(path);
      if(para.debug){
        cout<<"New one: ";
        for(int i=0;i<path.size();i++) cout<<path[i]<<" ";
        cout<<endl;
      }
    }else{
      isvisited[c]=0;
      for(int i=children.size()-1;i>=0;i--){
        //smallest comes first, so in a stack, it will iterate last
        int nextid=children[i];
        cstack.push(nextid);
        if(para.debug) cout<<"--Pushing "<<nextid<<endl;
      }
    }//end else
  }//end while
  return isvisited[start];
}


//void fillRCount(mi& rcount, mi& tpm, vi& tcsum){
//  for(int i=0;i<tpm.size();i++){
//    if(tpm[i].size()>0){
//      if(tpm[i].size()==1){
//        rcount[tpm[i][0]][tpm[i][0]]+=tcsum[i];
//      }else{
//        for(int j=0;j<tpm[i].size()-1;j++){
//          rcount[tpm[i][j]][tpm[i][j+1]]+=tcsum[i];
//        }
//      }
//    }
//  }//end for
//}

int enumiso(MInst& inst, mi & res,int maxiso){ 
  int nsegs=inst.getNSegs();
  mi rcount(nsegs,vi(nsegs,0));
  inst.fillRCount(rcount,0);

  //enumerate
  vi needcover(nsegs,1);

  int start=0;

  EnumParameter para;
  
  long k=0;
  while(para.minjuncfrac<1){ 
    res.clear();
    para.debug=true;
    para.immreturn=true;
    para.numonly=true;
    long k2=goenum_numonly(inst, rcount, start,res,  para);
    cout<<"Total:"<<k2<<", frac:"<<para.minjuncfrac<<endl;
    if(k2<=maxiso)break;
    para.minjuncfrac*=1.1;
  }

  para.immreturn=false;
  para.numonly=false;
  k=goenum_numonly(inst, rcount, start, res, para);
  cout<<"Total:"<<k<<", frac:"<<para.minjuncfrac<<endl;
  
  return 0;
}

/* Enumerate isoforms, and then calculate the expresion levels.
   Isoforms and their expression levels are stored in res and explv.
*/
int enumandcalexp(MInst& inst, mi& res, md& explv){
  //the old enumeration 
  enumiso(inst,res,10);
  //predict expression level
  cout<<"# of predicted isoforms: "<<res.size()<<endl;
  calcexplv(inst, res, explv);
  return 0;
}

